# -*- coding: UTF-8 -*-
##############################################################################
# Copyright (c) 2011 Timothy W. Cook and Contributors.
# All Rights Reserved. See CONTRIBUTORS.txt
#
# This software is subject to the provisions of the GNU LESSER GENERAL 
# PUBLIC LICENSE Version 3, Dated: 29 June 2007 See docs/LICENSE.txt or 
# http://www.gnu.org/licenses/lgpl.html 
##############################################################################

from zope.interface import Interface, Attribute
from zope.schema import Bool, TextLine, Object
from zope.i18nmessageid.message import MessageFactory

from mlhim.oship.rm.support.interfaces import IExceptionalValue

_ = MessageFactory('mlhim.rm')

class IDvAny(Interface):
    """
    Serves as a common ancestor of all data value types in MLHIM models. 
    """
    
    vtb = TextLine(
        description = _(u'Valid Time Begin - If present this must be a valid datetime string including timezone.'),
        title = _(u'Valid From'),
        required = False,
        default = '',
    )
    vte = TextLine(
        description = _(u'Valid Time End - If present this must be a valid datetime string including timezone'),
        title = _(u'Valid Until'),
        required = False,
        default = '',
    )
    ev = Object(
        description = _(u'The exceptional value. Often referred to as Null Flavour.'),
        title = _(u'Exceptional Value'),
        schema = IExceptionalValue,
        required = False,
        default = '',
    )

class IDvBoolean(IDvAny):
    """
    Items which are truly boolean data, such as true/false or yes/no answers.
    The invariant defined in the spec for this class is that it is not void.
    In Python a 'None' is defined as False.
    """

    dv = TextLine(
        title = _(u"Data Value"),
        description = _(u"""A string indicating a boolean type result according to the rules described below.
                        Boolean True values are represented by any one of these strings:  ["True","1","T"] 
                        Boolean False values are represented by any one of these strings:  ["False","0","F"] 
                        Note that the strings are not case-sensitive; i.e. "T" == "t" and "TRUE" == "t"
                        The empty string; "" represents a Void or Null value, in which case the ev attribute 
                        cannot be empty."""),
        required = False,
        default = False,
    )

class IDvText(IDvAny):
    """
    A text item, which may contain any amount of legal characters arranged as words, sentences etc. 
    """

    dv = TextLine(
        title = _(u"Data Value"),
        description = _(u"""Displayable rendition of the item, regardless of its underlying structure."""),
        required=False,
        default = '',
    )

    language = TextLine(
        title = _(u"Language"),
        description = _(u"""Optional indicator of the localised language in which the value is written. 
                        Only used when the text object is in a different language from the enclosing ENTRY."""),
        required = False,
        default = '',
        )


class IDvCodedText(IDvText):
    """
    A text item whose 'dv' must be the rubric from a controlled terminology, the
    key (i.e. the 'code') of which is the code_string attribute.
    Since DvCodedText is a subtype of DvText, it can be used in place of it,
    effectively allowing the type DvText to mean "a text item, which is coded".
    """

    code_string = TextLine(
        title = _(u"DefiningCode"),
        description = _(u"""The term which the 'value' attribute is the defining_code:CODE_PHRASE textual rendition (i.e. rubric) of."""),
        required = False,
        default = '',
        )

    terminology_name = TextLine(
        title = _(u"Terminology Name"),
        description = _(u"""Full Source Name from NLM Metathesarus; or similar."""),
        required = False,
        default = '',
        )

    terminology_abbrev = TextLine(
        title = _(u"Terminology Abbreviation"),
        description = _(u"""Version Source Abbreviation (VSAB) from NLM Metathesarus; or similar."""),
        required = False,
        default = '',
        )

class IDvParagraph(IDvAny):
    """
    A logical composite text value consisting of a series of DvText items.
    """

    items = List(
        value_type=Object(schema=IDvText),
        title = _(u"Items"),
        description = _(u"""Items making up the paragraph."""),
        required = False,
        default = [],
        )


class IDvIdentifier(IDvText):
    """
    Type for representing identifiers of real-world entities. Typical identifiers include
    drivers licence number, social security number, veterans affairs number, prescrip-
    tion id, order id, and so on. The actual identifier is in the 'dv' attribute.

    Used to represent any identifier of a real thing, issued by some authority or agency.
    """

    issuer = TextLine(
        title = _(u"Issuer"),
        description = _(u"""Authority which issues the kind of id used in the id field
                      of this object."""),
        required = False,
        default = '',
        )

    assigner = TextLine(
        title = _(u"Assigner"),
        description = _(u"""Organisation that assigned the id to the item being identified."""),
        required = False,
        default = '',
        )

    name = TextLine(
        title = _(u"Name"),
        description = _(u"""The identifier type, such as "prescription",or "SSN".
                      One day a controlled vocabulary might be possible for this."""),
        required = False,
        default = '',
        )


class IDvURI(IDvAny):
    """A reference to an object which conforms to the Universal Resource Identifier
       (URI) standard, as defined by W3C RFC 2936. """

    dv = URI(
        title = _(u"URI"),
        description = _(u"""A valid URI."""),
        required=False,
    )
   

    def scheme():
        """A distributed information "space" in which information objects exist. The scheme
        simultaneously specifies an information space and a mechanism for accessing objects
        in that space. For example if scheme = "ftp", it identifies the information space in
        which all ftpable objects exist, and also the application - ftp - which can be used
        to access them. Values may include: "ftp", "telnet", "mailto", "gopher" and many others.
        Refer to WWW URI RFC for a full list. New information spaces can be accommodated within
        the URI specification."""

    def path():
        """A string whose format is a function of the scheme. Identifies the location in
        <scheme>-space of an information entity. Typical values include hierarchical directory
        paths for any machine. For example, with scheme = "ftp", path might be /pub/images/image_01.
        The strings "." and ".." are reserved for use in the path. Paths may include internet/intranet
        location identifiers of the form: sub_domain...domain, e.g. "info.cern.ch" """

    def fragmentId():
        """A part of, a fragment or a sub-function within an object. Allows references to sub-parts
        of objects, such as a certain line and character position in a text object. The syntax and
        semantics are defined by the application responsible for the object. """


    def query():
        """Query string to send to application implied by scheme and path Enables queries to
        applications, including databases to be included in the URI Any query meaningful to the
        server, including SQL."""


class IDvDateTime(IDvQuantified):
    """
    Represents an absolute point in time, specified to the second. Semantics defined by ISO 8601.
    Used for recording a precise point in real world time, and for approximate time
    stamps, e.g. the origin of a HISTORY in an OBSERVATION which is only partially known.
    """

    value = TextLine(
        title=_(u"Value"),
        description=_(u"""ISO8601 date/time string"""),

        )

    def diff(other):
        """Difference of two date/times. Returns a DvDuration"""



    def valueValid():
        """validIso8601DateTime(value)"""
        
class IDvEncapsulated(IDvAny):
    """Abstract class defining the common meta-data of all types of encapsulated data."""

    size = Int(
        title=_(u"Size"),
        description=_(u"""Original size in bytes of unencoded encapsulated data."""),
        required=False,
        )

    charset = TextLine(
        title=_(u"charset"),
        description=_(u"""Name of character encoding scheme in which this value is encoded."""),
        required=False,
        )

    language = TextLine(
        title=_(u"Language"),
        description=_(u"""Optional indicator of the localised language in which the data is written"""),
        required=False,
        )

    def asString():
        """Result = alternate_text [(uri)]"""


class IDvMultimedia(IDvEncapsulated):
    """
    A specialisation of DvEncapsulated for audiovisual and biosignal types. Includes further
    metadata relating to multimedia types which are not applicable to other subtypes of DvEncapsulated.
    """
    
    alt_text = TextLine(
        title=_(u"Alternate Text"),
        description=_(u"""Text to display in lieu of multimedia display/replay"""),
        required=False,
        )

    mime_type = TextLine(
        title=_(u"Media Type"),
        description=_(u"""Data media type coded from openEHR code set "media types"
        (interface for the IANA MIME types code set)."""),

        )

    compression_algorithm =TextLine(
        title=_(u"Compression Algorithm"),
        description=_(u"""Compression type, a coded value from the openEHR "Integrity check"
        code set. Void means no compression."""),
        required=False,
        )

    integrity_check = BytesLine(
        title=_(u"Integrity Check"),
        description=_(u"""Binary cryptographic integrity checksum. Type==Array<Octet>"""),
        required=False,
        )

    integrity_check_algorithm = TextLine(
        title=_(u"Integrity Check Algorithm"),
        description=_(u"""Type of integrity check, a coded value from the openEHR "Integrity check" code set."""),
        required=False,
        )

    thumbnail = Field(
        title=_(u"Thumbnail"),
        required=False
    )

    uri = Object(
        schema=IDvURI,
        title=_(u"URI"),
        description=_(u"""URI reference to electronic information stored outside the record as a file,
        database entry etc, if supplied as a reference."""),
        required=False,
        )

    dv = BytesLine(
        title=_(u"Data"),
        description=_(u"""The actual data."""),
        required=False,
        )

    def isExternal(self):
        u"""Computed from the value of the uri attribute: True if the data is stored externally
        to the record, as indicated by 'uri'. A copy may also be stored internally, in which case
        'isExpanded' is also true.  Ensure uri !=None and uri != '' """

    def isInline(self):
        u"""Computed from the value of the data attribute: True if the data is stored in expanded
        form, ie within the EHR itself. Ensure data is not None. """

    def isCompressed(self):
        u"""Computed from the value of the compression_algorithm attribute: True if the data is
        stored in compressed form. Ensure compressionAlgorithm is not None. """

    def hasIntegrityCheck(self):
        u"""Computed from the value of the integrityCheckAlgorithm attribute: True if an
        integrity check has been computed. Ensure integrityCheckAlgorithm is not None."""

class IDvParsable(IDvEncapsulated):
    """
    Encapsulated data expressed as a parsable String. The internal model of the data item is not
    described in the openEHR model in common with other encapsulated types, but in this case, the
    form of the data is assumed to be plaintext, rather than compressed or other types of large binary data.
    Used for representing values which are formal textual representations, e.g. guidelines.
    """

    dv = TextLine(
        title=_(u"Value"),
        description=_(u"""The string, which may validly be empty in some syntaxes"""),

        )

    formalism = TextLine(
        title=_(u"Formalism"),
        description=_(u"""The name of the formalism, e.g. "GLIF 1.0", "proforma" etc."""),

        )

    def asString():
        u"""
        Result = alternate_text [(uri)]
        """


class IDvInterval(IDvAny):
    """
    Used to define intervals of dates, times, quantities (whose units match) and so on.
    The type parameter, T, must be a descendant of the type DV_ORDERED, which is
    necessary (but not sufficient) for instances to be compared (strictly_comparable
    is also needed).
    Without the DV_INTERVAL class, quite a few more DV_ classes would be needed
    to express logical intervals, namely interval versions of all the date/time classes,
    and of quantity classes. Further, it allows the semantics of intervals to be stated in
    one place unequivocally, including the conditions for strict comparison.
    The basic semantics are derived from the class INTERVAL<T>, described in the
    support RM.
    """
    
    lower=Object(
        schema=IDvOrdered,
        title=_(u"lower"),
        description=_(u"""The lower end of the interval."""),
        required=False
        )
    upper=Object(
        schema=IDvOrdered,
        title=_(u"upper"),
        description=_(u"""The upper end of the interval."""),
        required=False
        )
    lower_included=Bool(
        title=_(u"lower included"),
        description=_(u"""Is the lower end of the interval included."""),
        default=False
        )
    upper=Bool(
        title=_(u"upper included"),
        description=_(u"""Is the upper end of the interval included."""),
        default=False
        )


class IReferenceRange(Interface):
    """
    Defines a named range to be associated with any ORDERED datum. Each such
    range is particular to the patient and context, e.g. sex, age, and any other factor
    which affects ranges.
    May be used to represent normal, therapeutic, dangerous, critical etc ranges.
    """

    meaning=Object(
        schema=IDvText,
        title=_(u"meaning"),
        description=_(u"""Term whose value indicates the meaning of this range,  e.g. "normal", "critical", "therapeutic" etc."""),
        required=True
        )

    data_range=Object(
        schema=IDvInterval,
        title=_(u"Data Range"),
        description=_(u"""The data range for this meaning."""),
        required=True
        )

    def isInRange(val):
        """
        Indicates if the value 'val' is inside the range.
        """


class IDvOrdered(IDataValue):
    """
    Purpose:
    Abstract class defining the concept of ordered values, which includes ordinals as
    well as true quantities. It defines the functions to Arithmetic comparison
    such as '<' operator function  and is_strictly_comparable_to,
    the latter of which must evaluate to True for instances being compared with the
    '<' function, or used as limits in the DV_INTERVAL<T> class.

    Use:
    Data value types which are to be used as limits in the DV_INTERVAL<T> class
    must inherit from this class, and implement the function
    is_strictly_comparable_to to ensure that instances compare meaningfully. For
    example, instances of DV_QUANTITY can only be compared if they measure the
    same kind of physical quantity.
    """


    normal_range = Object(
        schema=IDvInterval,
        title = _(u"Normal Range"),
        description = _(u"Optional normal range."),
        required = False
        )

    other_reference_ranges = List(
        value_type=Object(schema=IReferenceRange),
        title = _(u"Other Reference Ranges"),
        description = _(u"""Optional tagged other reference ranges for this value in its particular measurement context. A list of ReferenceRange types."""),
        required = False
        )

    normal_status = Object(
        schema=ICodePhrase,
        title = _(u"Normal Status"),
        description = _(u"""Optional normal status indicator of value with
            respect to normal range for this value. Often included by lab, even
            if the normal range itself is not included. Coded by ordinals, 
            according to the openEHR terminology group "normal status"."""),
        required = False
    )


    def is_normal():
        """
        Value is in the normal range, determined by comparison of the
        value to the normalRange if present, or by the normalStatus
        marker if present.
        """

    def is_simple():
        """
        is_simple: Boolean
        True if this quantity has no reference ranges.
        """

    def is_strictly_comparable_to(other):
        """
        Test if two instances are strictly comparable.
        """

    @invariant
    def other_reference_ranges_validity(obj):
        if obj.other_reference_ranges is not None:
            if len(obj.other_reference_ranges) == 0:
                raise Invalid("The other_reference_ranges cannot be empty.")

    @invariant
    def is_simple_validity(obj):
        void_check = obj.normal_range is None and obj.other_reference_ranges is None
        if void_check:
                if not obj.is_simple():
                    raise Invalid("There is a inconsistency: the is_simple method cannot validate correctly.")

    @invariant
    def normal_range_status_consistency(obj):
        if obj.normal_range is not None and obj.normal_status is not None:
            if not xor(obj.normal_status.codeString == u"N" , obj.normal_range.has(obj)):
                raise Invalid(("Both normal_status and normal_range are defined"
                        " and validated correctly. Only one of them can be"
                        " validated by the time."))

    @invariant
    def normal_status_validity(self):
        #TODO: Implement the Terminology Service is mandatory to
        #implement this invariant
        pass


class IDvQuantified(IDvOrdered):
    """
    Abstract class defining the concept of true quantified values, i.e. values which are
    not only ordered, but which have a precise magnitude.
    """

    magnitude = Field(
        title=_(u"Magnitude"),
        description=_(u"""Numeric value of the quantity in canonical (i.e. single value)
                          form. Implemented as constant, function or attribute in subtypes as
                          appropriate. The type OrdereNumeric is mapped to the available
                          appropriate type in each implementation technology."""),
        required=True
    )

    accuracy = Attribute(_(u"Accuracy"))

    magnitude_status = TextLine(
        title=_(u"Magnitude status"),
        description=_(u"""Optional status of magnitude with values:
                                = : magnitude is a point value
                                < : value is < magnitude
                                > : value is > magnitude
                                <= : value is <= magnitude
                                >= : value is >= magnitude
                                ~ : value is approximately magnitude
                                If not present, meaning is =. """),
        required=False
    )

    error = Float(
        title=_(u"Error"),
        description=_((u"Accuracy of measurement, expressed either as a "
                       "half-range percent value or a half-range quantity.")),
        required=False
    )

    accuracy_is_percent = Bool(
        title=_(u"Percent"),
        description=_((u"Indicates that accuracy was recorded "
                        "as a percent value or as an absolute "
                        "quantity value.")),
        required=False
    )

 
    def valid_magnitude_status(val):
        """
        Test whether a string 'val' is one of the valid values for the magnitude_status attribute.
        """

    def accuracy_unknown():
        """
        True if accuracy is not known.
        """
    @invariant
    def magnitude_exists(obj):
        if obj.magnitude is None:
            raise Invalid("The magnitude attribute cannot be None.")

    @invariant
    def valid_status_valid(obj):
        if obj.magnitude_status is not None:
            if not obj.valid_magnitude_status(obj.magnitude_status):
                raise Invalid("The magnitude is not valid.")

class IDvOrdinal(IDvOrdered):
    """
    Models rankings and scores, e.g. pain, Apgar values, etc, where there is a)
    implied ordering, b) no implication that the distance between each value is con-
    stant, and c) the total number of values is finite.

    Used for recording any clinical datum which is customarily recorded using sym-
    bolic values. Example: the results on a urinalysis strip, e.g. {neg, trace, +,
    ++, +++} are used for leucocytes, protein, nitrites etc; for non-haemolysed
    blood {neg, trace, moderate}; for haemolysed blood {neg, trace,
    small, moderate, large}.
    """

    dv = Int(
        title=_(u"value"),
        description=_(u""" Ordinal position in enumeration of values. """),
        required=True
    )

    symbol = TextLine(
        title=_(u"symbol"),
        description=_(u"""Textual representation of this
                       value in the enumeration, which may be strings made from "+" symbols,
                       or other enumerations of terms such as "mild", "moderate", "severe",
                       or even the same number series as the values,
                       e.g. "1", "2", "3". Codes come from archetype."""),
        required=True
    )


    def limits(ref_range):
        """
        limits of the ordinal enumeration, to allow
        comparison of an ordinal value to its limits.
        Returns DvOrdinal.
        """


        
class IDvDuration(IDvAmount):
    """
    Represents a period of time with respect to a notional point in time, which is not
    specified. A sign may be used to indicate the duration is "backwards" in time
    rather than forwards.

    Note that a deviation from ISO8601 is supported, allowing the 'W' designator to
    be mixed with other designators. See assumed types section in the Support IM.

    Used for recording the duration of something in the real world, particularly when
    there is a need a) to represent the duration in customary format, i.e. days, hours,
    minutes etc, and b) if it will be used in computational operations with date/time
    quantities, i.e. additions, subtractions etc.
    """

    dv = TextLine(
        title=_(u"Value"),
        description=_(u"""ISO8601 duration"""),
        required=True,
        )

    def magnitude():
        """
        Numeric value of the duration in seconds.
        Result >= 0.0
        """

    @invariant
    def value_valid(obj):
        if not obj.valid_ISO8601_duration(obj.value):
            raise Invalid("Invalid ISO8601 format.")


class IDvCount(Interface):
    """
    Purpose: Countable quantities

             Used for countable types such as pregnancies and steps (taken by a physiotherapy
    Use:     patient), number of cigarettes smoked in a day.

    Misuse:  Not used for amounts of physical entities (which all have units)
    """

    magnitude = Int(
        title=_(u"Magnitude"),
        description=_(u"""numeric magnitude of the quantity"""),
        required=True
        )


class IDvProportion(Interface):
    """
              Models a ratio of values, i.e. where the numerator and denominator are both pure
    Purpose:  numbers.

              Used for recording titers (e.g. 1:128), concentration ratios, e.g. Na:K (unitary
      Use:    denominator), albumin:creatinine ratio, and percentages, e.g. red cell distirbution
              width (RDW).

              Should not be used to represent things like blood pressure which are often written
              using a / character, giving the misleading impression that the item is a ratio,
    MisUse:   when in fact it is a structured value. E.g. visual acuity 6/24 is not a ratio.
              Should not be used for formulations.
    """


    numerator = Float(
        title=_(u"numerator"),
        description=_(u"""numerator of ratio"""),
        required=True
    )

    denominator = Float(
        title=_(u"denominator"),
        description=_(u"""denominator of ratio"""),
        required=True
    )

    proportion_type = Int(
        title=_(u"type"),
        description=_(u"""Indicates semantic type of proportion:
                        pk_ratio = Ratio type. Numerator and denominator may be any value. Constant value = 0
                        pk_unitary	= Denominator must be 1. Constant value = 1
                        pk_percent	= Denominator is 100, numerator is understood as a percentage value. Constant value = 2
                        pk_fraction	=	Numerator and denominator are integral, and the presentation method uses a slash, 
                            e.g. “1/2”. Constant value = 3
                        pk_integer_fraction	 = 	Numerator and denominator are integral, and the presentation method uses a 
                            slash, e.g. “1/2”; if the numerator is greater than the denominator, e.g. n=3, d=2, the 
                            presentation is “1 1/2”. Constant value = 4
                    """),
        required=True
    )

    precision = Int(
        title=_(u"precision"),
        description=_(u"""Precision to which the numerator and denominator values of the proportion
                    are expressed, in terms of number of decimal places. The value 0 implies an
                    integral quantity. The value -1 implies no limit, i.e.any number of decimal places."""),
        required=False
    )


    def isIntegral():
        """
        True if the numerator and denominator values are integers, i.e. if the precision is 0.
        """

    def magnitude():
        """
        Effective magnitude represented by ratio.
        Result = numerator / denominator
        """


class IDvQuantity(Interface):
    """
    Quantitified type representing scientific quantities, i.e. quantities expressed as a
    magnitude and units.
    Units were inspired by the Unified Code for Units of Measure (UCUM), devel-
    oped by Gunther Schadow and Clement J. McDonald of The Regenstrief Institute.

    Can also be used for time durations, where it is more convenient to treat these as
    simply a number of seconds rather than days, months, years.
    """

    magnitude = Float(
        title=_(u"Magnitude"),
        description=_(u"""Numeric magnitude of the quantity."""),
        required=True
        )

    units = TextLine(
        title=_(u"Units"),
        description=_(u"""Stringified units, expressed in UCUM unit syntax,
                    e.g. "kg/m2", "mm[Hg]", "ms-1", "km/h".Implemented accordingly in subtypes."""),

        )


    precision = Int(
        title=_(u"Precision"),
        description=_(u"""Precision to which the value of the quantity is expressed, in terms of
                    number of decimal places. The value 0 implies an integral quantity.
                    The value -1 implies no limit, i.e. any number of decimal places."""),
        required=False,
        )

    def isIntegral():
        """True if precision = 0; quantity represents an integral number."""


    def isStrictlyComparableTo(other):
        """Test if two instances are strictly comparable by ensuring that the measured
        property is the same, achieved using the Measurement service function units_equivalent.

        Return selfunits == other.units
        """


class IDvTimeSpecification(Interface):
    """
    This is an abstract class of which all timing specifications are specialisations.
    Specifies points in time, possibly linked to the calendar, or a real world repeating
    event, such as "breakfast".
    """

    value = Attribute("""the specification, in the HL7v3 syntax for PIVL or EIVL types.
                    See section 8.2.2.1 Phase-linked Time Specification Syntax""")

    def calendarAlignment():
        """Indicates what prototypical point in the calendar the specification is
        aligned to, e.g. "5th of the month". Empty if not aligned. Extracted from
        the 'value' attribute.
        """

    def eventAlignment():
        """Indicates what real-world event the specification is aligned to if any.
        Extracted from the 'value' attribute.
        """

    def institutionSpecified():
        """Indicates if the specification is aligned with institution schedules,
        e.g. a hospital nursing changeover or meal serving times. Extracted from
        the 'value' attribute.
        """
     


class IDvGeneralTimeSpecification(Interface):
    """Specifies points in time in a general syntax. Based on the HL7v3 GTS data type."""

    def calendarAlignment():
        """Calendar alignment extracted from value. """

    def eventAlignment():
        """Event alignment extracted from value."""

    def institutionSpecified():
        """Extracted from value."""

    def valueValid():
        """value.formalism.is_equal("HL7:GTS")"""


class IDvPeriodicTimeSpecification(Interface):
    """
    Specifies periodic points in time, linked to the calendar (phase-linked),
    or a real world repeating event, such as "breakfast" (event-linked).
    Based on the HL7v3 data types PIVL<T> and EIVL<T>.
    Used in therapeutic prescriptions, expressed as INSTRUCTIONs in the openEHR model.
    """

    def period():
        """The period of the repetition, computationally derived from the syntax
        representation. Extracted from the 'value' attribute. Returns a DvDuration.
        """

    def calendarAlignment():
        """Calendar alignment extracted from value."""


    def eventAlignment():
        u"""Event alignment extracted from value."""

    def institutionSpecified():
        """Extracted from value. """

