# -*- coding: UTF-8 -*-
##############################################################################
# Copyright (c) 2011 Timothy W. Cook and Contributors.
# All Rights Reserved. See CONTRIBUTORS.txt
#
# This software is subject to the provisions of the GNU LESSER GENERAL 
# PUBLIC LICENSE Version 3, Dated: 29 June 2007 See docs/LICENSE.txt or 
# http://www.gnu.org/licenses/lgpl.html 
##############################################################################

from zope.interface import Interface, Attribute
from zope.schema import Bool, TextLine, Object
from zope.i18nmessageid.message import MessageFactory

_ = MessageFactory('mlhim.rm')

class IDataStructure(Interface):
    u"""
     Abstract parent class of all data structure types. Includes the
     as_hierarchy function which can generate the equivalent CEN EN13606 single
     hierarchy for each subtype's physical representation. For example, the
     physical representation of an ITEM_LIST is List<ELEMENT>; its
     implementation of as_hierarchy will generate a CLUSTER containing the set
     of ELEMENT nodes from the list. """

    def asHierarchy():
        u"""Hierarchical equivalent of the physical representation of each
        subtype, compatible with CEN EN 13606 structures. Returns a List."""

        
class IItemStructure(Interface):
    u"""
    Abstract parent class of all spatial data types.
    """
    pass


class IItemList(Interface):
    u"""
    Logical list data structure, where each item has a value and can be
    referred to by a name and a positional index in the list.
    The list may be empty. """

    items = List(
        value_type=Object(schema=IElement),
        title=_(u"items"),
        description=_(u"""Physical representation of the list."""),
        required=False)

    def itemCount():
        u"""Count of all items."""

    def names():
        u"""Return the names of all items."""

    def namedItem(str):
        u"""Return the named 'str'"""

    def ithItem(n):
        u"""Return the 'n' ith item."""

    def asHierarchy():
        u"Generate a CEN EN13606-compatible hierarchy consisting of a"
        " single CLUSTER containing the ELEMENTs of this list."



class IItemSingle(Interface):
    u"""
    Logical single value data structure.
    Used to represent any data which is logically a single value, such as a
    person's height or weight. """

    item = Object(
        schema=IElement,
        title=_(u"item"),
        description=_(u"""Single item."""),
        required=True)

    def asHierarchy():
        u"""
        Generate a CEN EN13606-compatible hierarchy consisting of a single
        ELEMENT. """


class IItemTable(Interface):
    u"""
        Logical relational database style table data structure, in which
        columns are named and ordered with respect to each other. Implemented
        using Cluster-per-row encoding. Each row Cluster must have an identical
        number of Elements, each of which in turn must have identical names and
        value types in the corresponding postions in each row. Some columns may
        be designated 'key' columns, containing key data for each row, in the
        manner of relational tables. This allows row-naming, where each row
        represents a body site, a blood antigen etc. All values in a column
        have the same data type. Used to represent any data which is logically
        a table of values, such as blood pressure, most protocols, many blood
        tests etc. Not used for time-based data, which should be represented
        with the temporal class HISTORY.. The table may be empty.
    """

    rows = List(
        value_type=Object(schema=ICluster),
        title=_(u"rows"),
        description=_(u"""Physical representation of the table as a list of
        CLUSTERs, each containing the data of one row of the table."""),
        required=False)

    def rowCount():
        u"""Return number of rows as an integer."""

    def columnCount():
        u"""Return the number of columns as an integer."""

    def rowNames():
        u"""Return the List of row names as DvText items."""

    def columnNames():
        u"""Return the List of column names as DvText items."""

    def ithRow(n):
        u"""Return the 'n' ith row as a Cluster."""

    def hasRowWithName(a_key):
        u"""Return True if a_key exists as a row name."""

    def hasColumnWithName(a_key):
        u"""Return True if a_key exists as a column name."""

    def namedRow(a_key):
        u"""Return the row whose first column has the name a_key."""

    def hasRowWithKeys(keys):
        u"""Return True if a row's first n columns contain the names of set of
        keys."""

    def rowWithKeys(keys):
        u"""Return the row whose first n columns contain the names in the set
        of keys."""

    def elementAtCell(i, j):
        u"""
        NOTE: In the specs this is called elementAtCellij. I think it should be
        changed. Return the element at the intersection of column i and row j.
        i and j are integers. """

    def elementAtNamedCell(row_key, col_key):
        u"""Return the element at the intersection or col_key and row_key """

    def asHierarchy():
        u"""Generate a CEN EN13606-compatible hierarchy consisting of a single
        CLUSTER containing the ELEMENTs of this list."""


class IItemTree(Interface):
    u"""
    Logical tree data structure.
    """

    # the list contents should be restricted to schema = IItem,
    items = List(
        value_type=Object(schema=IItem),
        title=_(u"items"),
        description=_(u"Physical representation of the tree."),
        required=False)

    def hasElementPath(a_path):
        u"""Return True if a_path is a valid leaf element path."""

    def elementAtPath(a_path):
        u"""Return the leaf element at the path; a_path."""

    def asHierarchy():
        u"""Generate a CEN EN13606-compatible hierarchy consisting of a single
        CLUSTER containing the ELEMENTs of this list."""
class IItem(Interface):
    u"""
    The abstract parent of CLUSTER and ELEMENT representation classes.
    """
    pass


class IElement(Interface):
    u"The leaf variant of ITEM, to which a DATA_VALUE instance is attached."

    value = Object(
        schema=IDataValue,
        title=_(u"value"),
        description=_(u"""Data value of this leaf."""),
        required=False)

    nullFlavour = Object(
        schema=IDvCodedText,
        title=_(u"""nullFlavour"""),
        description=_(u"Flavour of null value, e.g. 'indeterminate',"
                      " 'not asked', etc."),
        required=False)

    def isNull():
        u"""Return True if value is unknown, etc."""

    def nullFlavourValid(obj):
        u"If value is None then nullFlavour must be in the terminology"
        " code set for null Flavours."


class ICluster(Interface):
    u"""
    The grouping variant of ITEM, which may contain further instances of ITEM,
    in an ordered list. """

    # the list contents should be restricted to schema = IItem,
    items = List(
        value_type=Object(schema=IItem),
        title=_(u"items"),
        description=_(u"Ordered list of items - CLUSTER or ELEMENT objects"
                      " - under this CLUSTER."),
        required=True)


class IEvent(Interface):
    u"""
    Defines the abstract notion of a single event in a series. This class is
    generic,allowing types to be generated which are locked to particular
    spatial types, such as EVENT<ITEM_LIST> Subtypes express point or interval
    data. """

    time=Object(
        schema=IDvDateTime,
        title=_(u'time'),
        description=_(u"""Time of this event. If the width is non-zero, it is
        the time point of the trailing edge of the event."""),
        required=True)


    data=Object(
        schema=IItemStructure,
        title=_(u'data'),
        description=_(u'The data of this event.'),
        required=True)

    state=Object(
        schema=IItemStructure,
        title=_(u'state'),
        description=_(u'Optional state information for this event.'),
        required=False)

    def parent():
        u"""
        Parent IHistory where the Event is Contained.
        It's redefinition of LOCATABLE.parent to be of type History
        """

    def offset():
        u"""
        Offset of this event from origin , computed as
        time.diff(parent.origin) .
        """


class IEventContainer(IContainer):
    contains(IEvent)


class IHistory(Interface):
    u"""
    Root object of a linear history, i.e. time series structure. For a periodic
    series of events, period will be set, and the time of each Event in the
    History must correspond; i.e. the EVENT.offset must be a multiple of period
    for each Event. Missing events in a period History are however allowed.

    NOTE: The invariants have NOT been written yet for this interface.

    """

    origin = Object(
        schema=IDvDateTime,
        title=_(u"origin"),
        description=_(u"Time origin of this event history. The first event is"
                        " not necessarily at the origin point."),
        required=True)

    events = Object(
        schema=IEventContainer,
        title=_(u"events"),
        description=_(u"The events in the series."),
        required = False,
        )

    period=Object(
        schema=IDvDuration,
        title=_(u"period"),
        description=_(u"Period between samples in this segment if periodic."),
        required=False)

    duration=Object(
        schema=IDvDuration,
        title=_(u"duration"),
        description=_(u"Duration of the entire History; either corresponds to"
        " the duration of all the events, and/or the duration represented by"
        " the summary, if it exists."),
        required=False)

    summary=Object(
        schema=IItemStructure,
        title=_(u"summary"),
        description=_(u"Optional summary data expressing e.g. text or image"
        " which summarises entire History."),
        required=False)


class IEventContained(IContained):
    containers(IEventContainer)
    

class IIntervalEvent(Interface):
    u"""
    Defines a single interval event in a series.
    """

    width=Object(
        schema=IDvDuration,
        title=_(u"width"),
        description=_(u"Length of the interval during which the state was"
        " true."),
        required=True)

    mathFunction=Object(
        schema=IDvCodedText,
        title=_(u"mathFunction"),
        description=_(u'Mathematical function of the data of this event, e.g.'
        ' "maximum", "mean" etc. Coded using openEHR Terminology group'
        ' "event math function".'),
        required=True)

    sampleCount=Int(
        title=_(u"sampleCount"),
        description=_(u"Optional count of original samples to which this event"
                      " corresponds."),
        required=False)

    def intervalStartTime():
        u"""Start time of the interval of this event."""
