# -*- coding: UTF-8 -*-
##############################################################################
# Copyright (c) 2011 Timothy W. Cook and Contributors.
# All Rights Reserved. See CONTRIBUTORS.txt
#
# This software is subject to the provisions of the GNU LESSER GENERAL 
# PUBLIC LICENSE Version 3, Dated: 29 June 2007 See docs/LICENSE.txt or 
# http://www.gnu.org/licenses/lgpl.html 
##############################################################################

from zope.interface import Interface, Attribute
from zope.schema import Bool, TextLine, Object
from zope.i18nmessageid.message import MessageFactory

_ = MessageFactory('mlhim.rm')

class IExceptionalValue(Interface):
    """
    Serves as a common ancestor of all exceptional value types in MLHIM
    models. Replaces the concept of NullFlavours. 
    """

    pass


class IObjectRef(Interface):
    u""" Class describing a reference to another object, which may exist locally
    or be maintained outside the current namespace, e.g. in another service.
    Services are usually external, e.g. available in a LAN (including on the
    same host) or the inter- net via Corba, SOAP, or some other distributed
    protocol.  However, in small systems they may be part of the same executable
    as the data containing the Id.
    """

    id = Object(
        schema=IObjectId,
        title = u"Id",
        description = (u"Globally unique id of an object (of type ObjectId),"
                        "regardless of where it is stored."),
        required = True,
    )

    namespace = Text(
        title = u"Namespace",
	description = _((u"Namespace to which this identifier belongs in the"
                      "local system context.  ")),
        constraint=legal_namespaces,
        required = True,
    )

    type = Attribute(("Name of the class (concrete or abstract)"
                      "of object to which this identifier type"
                      "refers, e.g.PARTY, PERSON, GUIDELINE"
                      "etc.  These class names are from the relevant"
                      "reference model.  The type name ANY can be"
                      "used to indicate that any type is accepted"
                      "(e.g. if the type is unknown)."))


    @invariant
    def id_exists(objref):
        if objref.id is None or not len(objref.id.value):
            raise Invalid("A id must exists.")

    @invariant
    def namespace_exists(objref):
        if objref.namespace is None or not len(objref.namespace):
            raise Invalid("A namespace must exists.")

    @invariant
    def type_exists(objref):
        if objref.type is None or not len(objref.type):
            raise Invalid("A type must exists.")


   
class IEntityRef(IObjectRef):
    u""" Identifier for parties in a demographic or identity
    service. There are typically a number of subtypes of the PARTY
    class, including PERSON, ORGANISATION, etc.

    Abstract supertypes are allowed if the referenced object is of a
    type not known by the current implementation of this class (in
    other words, if the demographic model is changed by the addition
    of a new PARTY or ACTOR subtypes, valid PartyRefs can still be
    constructed to them).
    """

    type = TextLine(
        title = _(u"Type"),
        description = _(u"""Name of the class (concrete or abstract)
                        of object to which this identifier type
                        refers, e.g."PARTY", "PERSON", "GUIDELINE"
                        etc.  These class names are from the relevant
                        reference model.  The type name "ANY" can be
                        used to indicate that any type is accepted
                        (e.g. if the type is unknown)."""),
        constraint=partyref_type_is_valid,
        )
